MODULE PZConvertSyntheticCands EXPORTS Main;

(* Convert synthetic candidates from fraction format to ".can" format  *)
(* Last edited on 1999-11-03 18:10:45 by hcgl *)

IMPORT ParseParams, FGet, NGet, FileRd, FileWr, TextWr, Wr, Fmt, Rd;
IMPORT Process, OSError, Thread, Stdio;
IMPORT PZLR3Chain, PZCandidate, PZSegment;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, NAT, BOOLS;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;        (* Input candidate file name (without ".txt") *)
      chainDir: TEXT;     (* Directory where contour files reside. *)
      chainPrefix: TEXT;  (* Prefix for chain file names *)
      band: NAT;          (* Nominal band width (lambda) for file names *)
      step: LONG;         (* Nominal sampling step size. *)
      output: TEXT;       (* Output candidate file name (without ".can") *)
    END;

  FractionInterval = RECORD    (* Identifies a segment of a filtered curve *)
      cvx: NAT;     (* Curve identifier *)
      ini: LONG;    (* Segment beginning (as fraction of length *)
      fin: LONG;    (* Segment end (as fraction of length  *)
    END;

  FractionList = ARRAY OF ARRAY[0..1] OF FractionInterval;
  
  ChainList = ARRAY OF PZLR3Chain.ReadData;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions(),
      lambda =  FLOAT(o.band,LONG),
      rd = FileRd.Open(o.input & ".txt"),
      perc = ReadFractionFile(rd)^,
      curves = GetCurves(perc)^,
      chData = PZLR3Chain.ReadAll(
        o.chainPrefix, o.band, ".flc", sel := curves, dir := o.chainDir,
        headerOnly := TRUE
      ),
      cand = ConvertCandidates(perc, chData.chain^, o.step)^,
      wr = FileWr.Open(o.output & ".can")
    DO
      PZCandidate.Write(wr, Comments(o), cand, lambda);
    END;
  END Main;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();

        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;

        pp.getKeyword("-band");
        o.band := pp.getNextInt();

        pp.getKeyword("-step");
        o.step := pp.getNextLongReal(0.0d0);

        pp.getKeyword("-output");
        o.output := pp.getNext();

        pp.finish();
      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZConvertSyntheticCands \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  [ -chainDir NAME ] -chainPrefix STRING \\\n");
          Wr.PutText(stderr, "  -band NUMBER -step NUMBER \\\n");
          Wr.PutText(stderr, "  -output NAME \\\n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE GetCurves(READONLY perc: FractionList): REF BOOLS =
  CONST MaxCurves = 10000;
  VAR mxc: NAT := 0; (* Max curve number in "cand" *)
  BEGIN
    WITH seen = NEW(REF BOOLS, MaxCurves)^ DO
      FOR k := 0 TO LAST(seen) DO seen[k] := FALSE END;
      FOR i := 0 TO LAST(perc) DO
        FOR j := 0 TO 1 DO
          WITH k = perc[i][j].cvx DO
            IF NOT seen[k] THEN
              seen[k] := TRUE;  mxc := MAX(mxc, k)
            END;
          END
        END
      END;
      mxc := mxc+1;
      WITH rb = NEW(REF BOOLS, mxc) DO
        rb^ := SUBARRAY(seen, 0, mxc);
        RETURN rb
      END
    END
  END GetCurves;

PROCEDURE ConvertCandidates(
    READONLY perc: FractionList;
    READONLY ch: ChainList;
    step: LONG;
  ): REF PZCandidate.List =
  VAR segL, canL: LONG;
  BEGIN
    WITH
      n = NUMBER(perc),
      rc = NEW(REF PZCandidate.List, n),
      cand = rc^
    DO
      FOR i := 0 TO n-1 DO
        WITH ci = cand[i] DO
          canL := 0.0d0;
          FOR j:= 0 TO 1 DO
            WITH
              p = perc[i,j],
              m = ch[p.cvx].samples
            DO
              <* ASSERT m > 0 *>
              WITH
                ini = FLOOR(FLOAT(m, LONG) * p.ini),
                fin = CEILING(FLOAT(m, LONG) * p.fin),
                ns = (fin - ini) MOD m + 1
              DO
                ci.seg[j] := PZSegment.T{
                  cvx := p.cvx,
                  tot := m,
                  ini := ini,
                  ns := ns,
                  rev := (j = 1)
                };
                segL := step * FLOAT(ns - 1, LONG);
                canL := canL + segL;
              END
            END
          END;
          ci.length := canL/2.0d0;
          ci.mismatch := 0.0d0;
          ci.matchedLength := ci.length;
          ci.pm := NIL
        END
      END;
      RETURN rc
    END
  END ConvertCandidates;

PROCEDURE Comments(READONLY o: Options):TEXT=
  BEGIN
    WITH
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutText(wr, "PZConvertSyntheticCands: \n");
      Wr.PutText(wr, "  input:      " & o.input & "\n");
      Wr.PutText(wr, "  chainPrefix: " & o.chainPrefix     & "\n");
      Wr.PutText(wr, "  band:        " & Fmt.Int(o.band) & "\n");
      Wr.PutText(wr, "  output:     " & o.output & "\n");
      RETURN(TextWr.ToText(wr))
    END (* DO *);
  END Comments;

PROCEDURE ReadFractionFile(rd :Rd.T): REF FractionList  =
  VAR c: REF FractionList;
  BEGIN
    WITH
      n = NGet.Int(rd, "n")
     DO
      FGet.EOL(rd);
      c := NEW(REF FractionList, n);
      FOR i := 0 TO n-1 DO
        FGet.Skip(rd, FGet.Spaces);
        WITH ci = c[i] DO
          FOR j := 0 TO 1 DO
            WITH seg = ci[j] DO
              seg.cvx := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              seg.ini := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
              seg.fin := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
            END;
          END;
          FGet.EOL(rd)
        END
      END;
      RETURN c;
    END;
  END ReadFractionFile;

BEGIN
  Main()
END PZConvertSyntheticCands.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
