MODULE PZDrawCands EXPORTS Main;

(* Draws a list of candidate matchings   *)
(* Last edited on 1999-11-21 23:12:44 by hcgl *)

IMPORT PSPlot, LR3, LR4x4; 
IMPORT ParseParams, FileRd, Wr, Fmt;
IMPORT Process, OSError, Thread, Stdio;
IMPORT PZLR3Chain, PZCandidate;
IMPORT PZPlot, PZGeo;  

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;             (* Input candidate file name *)
      chainDir: TEXT;          (* Directory where chains can be found. *)
      chainPrefix: TEXT;       (* Chain file name prefix *)
      band: NAT;               (* Nominal band width (lambda) for file names *)
      extension: TEXT;         (* Extension for geometric chains (".flc" usually) *)
      output: TEXT;            (* Output plot file (without ".ps"/".eps") *)
      maxCands: NAT;           (* Number of candidates to plot. *)
      
      (* Plotting options *)
      drawEvery: NAT;          (* Plot only one sample every this many. *) 
      drawMatch: BOOL;         (* TRUE to draw the matching (if present). *)
      displace: LONG;          (* Separate chains by this amount when plotting *)
      noPointers: BOOL;        (* TRUE omits the arrowheads at the segment ends. *)
      noFrame: BOOL;           (* TRUE omits the frame around each candidate. *)
      epsFormat: BOOL;         (* TRUE to use Encapsulated Postscript. *)
      plotWidth: LONG;         (* Nominal candidate plot width in mm. *)
      plotHeight: LONG;        (* Nominal candidate plot width in mm. *)
      nCandsPerRow: NAT;       (* Number of candidate plots per row (1 for ".eps"). *)
      nCandsPerCol: NAT;       (* Number of candidate plots per column (1 for ".eps"). *)
    END;

PROCEDURE Main() =
  BEGIN
    WITH 
      o = GetOptions()
    DO
      IF o.maxCands > 0 THEN
        WITH
          candData = PZCandidate.Read(FileRd.Open(o.input & ".can")),
          cand = candData.c^,
          curves = PZCandidate.GetCurves(cand)^,
          chData = PZLR3Chain.ReadAll(
            o.chainPrefix, o.band, o.extension, sel := curves,
            dir := o.chainDir, headerOnly := FALSE
          )
        DO
          DrawCandidates(o.output, cand, chData.chain^, o);
        END;
      END;
    END;
 END Main;          

      
PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();
        
        IF pp.keywordPresent("-chainDir") THEN 
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;
        
        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();
        
        pp.getKeyword("-band");
        o.band := pp.getNextInt();
          
        IF pp.keywordPresent("-extension") THEN 
          o.extension:= pp.getNext();
        ELSE 
          o.extension := ".flc";
        END;
        
        pp.getKeyword("-output");
        o.output := pp.getNext();
               
        IF pp.keywordPresent("-maxCands") THEN
          o.maxCands := pp.getNextInt(1)
        ELSE
          o.maxCands := 160
        END;
        
        IF pp.keywordPresent("-drawEvery") THEN 
          o.drawEvery:= pp.getNextInt(1);
        ELSE 
          o.drawEvery := 1;
        END;
        
        o.drawMatch := pp.keywordPresent("-drawMatch");
        
        IF pp.keywordPresent("-displace") THEN
          o.displace := pp.getNextLongReal(0.0d0, LAST(LONG))
        ELSE
          o.displace := 0.0d0;
        END;
        
        o.epsFormat := pp.keywordPresent("-epsFormat");
        
        o.noPointers := pp.keywordPresent("-noPointers");
        
        o.noFrame := pp.keywordPresent("-noFrame");
        
        IF o.epsFormat THEN
          o.nCandsPerRow := 1;
          o.nCandsPerCol := 1;
          IF pp.keywordPresent("-plotSize") THEN 
            o.plotWidth  := pp.getNextLongReal(10.0d0, 1000.0d0);
            o.plotHeight := pp.getNextLongReal(10.0d0, 1000.0d0);
          ELSE 
            o.plotWidth  := 40.0d0;
            o.plotHeight := 40.0d0;
          END;
        ELSE
          IF pp.keywordPresent("-nCandsPerPage") THEN 
            o.nCandsPerRow:= pp.getNextInt(1, 100);
            o.nCandsPerCol:= pp.getNextInt(1, 100);
          ELSE 
            o.nCandsPerRow := 4;
            o.nCandsPerCol := 5;
          END;
          o.plotWidth  := 160.0d0/FLOAT(o.nCandsPerRow, LONG);
          o.plotHeight := 200.0d0/FLOAT(o.nCandsPerCol, LONG);
        END;
        
        pp.finish();   
        
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: PZDrawCands \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix NAME \\\n");
          Wr.PutText(stderr, "  -band NUMBER [ -extension EXT ] \\\n");
          Wr.PutText(stderr, "  -output NAME \\\n");
          Wr.PutText(stderr, "  [ -maxCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -displace NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -drawMatch ] [ -noPointers ] [ -noFrame ] \\\n");
          Wr.PutText(stderr, "  [ -epsFormat [ -plotSize WIDTH HEIGHT ] | \\\n");
          Wr.PutText(stderr, "    -nCandsPerPage NUMBER NUMBER \\\n");
          Wr.PutText(stderr, "  ] \\\n");
          Wr.PutText(stderr, "  [ -drawEvery NUMBER ] \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE DrawCandidates(
    output: TEXT;
    READONLY cand: PZCandidate.List;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  )=
  VAR ps: PSPlot.File;
  BEGIN 
    WITH 
      nCands = MIN(o.maxCands, NUMBER(cand)),
      nCandsPerPage = o.nCandsPerRow * o.nCandsPerCol,
      nPages = (nCands + nCandsPerPage - 1) DIV nCandsPerPage
    DO
      IF NOT o.epsFormat THEN
        WITH name = output & ".ps" DO
          ps := NEW(PSPlot.PSFile).open(name);
          Wr.PutText(stderr, "writing " & name & "\n");  
        END
      END;
    
      FOR page := 0 TO nPages-1 DO 
        WITH
          firstCandInPage = page*nCandsPerPage,
          lastCandInPage = MIN(nCands, (page+1)*nCandsPerPage) - 1,
          nCandsInPage = lastCandInPage - firstCandInPage + 1,
          caption = "Candidates [" & 
            Fmt.Int(firstCandInPage) & ".." & Fmt.Int(lastCandInPage) &
          "]"
        DO
          IF o.epsFormat THEN
            <* ASSERT firstCandInPage = lastCandInPage *>
            WITH 
              pg = Fmt.Pad(Fmt.Int(page), 5, '0'),
              name = output & "-pg" & pg & ".eps",
              xSize = o.plotWidth,
              ySize = o.plotHeight
            DO
              ps := NEW(PSPlot.EPSFile).open(name, xSize := xSize, ySize := ySize);
              Wr.PutText(stderr, "writing " & name & "\n");  
            END
          ELSE
            WITH ps = NARROW(ps, PSPlot.PSFile) DO
              ps.beginPage();
              (* ps.caption(output); *)
              (* ps.caption(caption); *) 
            END;
          END;
          Wr.PutText(stderr, caption & "\n");
          DrawPage(ps, SUBARRAY(cand, firstCandInPage, nCandsInPage), flc, o);
          IF o.epsFormat THEN 
            ps.close()
          ELSE
            WITH ps = NARROW(ps, PSPlot.PSFile) DO ps.endPage() END;
          END;
        END;
      END;
      IF NOT o.epsFormat THEN ps.close() END;
    END   
  END DrawCandidates;
 
PROCEDURE DrawPage(
    ps: PSPlot.File;
    READONLY cand: PZCandidate.List;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  )=
  CONST NominalCandidateWidth = 1000.0d0;
  BEGIN 
    FOR i := 0 TO LAST(cand)  DO 
      IF NOT o.epsFormat THEN 
        BeginDrawing(ps, o.nCandsPerRow, o.nCandsPerCol, o.plotWidth, o.plotHeight, i)
      END; 
      WITH 
        aspect = o.plotHeight/o.plotWidth,
        xMax = NominalCandidateWidth/2.0d0,
        yMax = NominalCandidateWidth/2.0d0*aspect
      DO
        ps.setScale(PSPlot.Axis.X, -xMax, +xMax);
        ps.setScale(PSPlot.Axis.Y, -yMax, +yMax);
      END;
      ps.setLineWidth(0.2);
      ps.setLineColor(PSPlot.Black);
      IF NOT o.noFrame THEN ps.frame() END;
      DrawCandidate(ps, cand[i], flc, o);
      IF NOT o.epsFormat THEN EndDrawing(ps) END;
    END   
  END DrawPage;


PROCEDURE BeginDrawing(
    ps: PSPlot.PSFile; 
    nCandsPerRow: NAT; 
    nCandsPerCol: NAT;
    plotWidth: LONG;
    plotHeight: LONG;
    i: NAT;
  ) =
  BEGIN
    WITH 
      totWidth = FLOAT(nCandsPerRow, LONG) * plotWidth,
      plotXMargin = 0.5d0 * (PSPlot.LetterXSize - totWidth),
      plotYMargin = plotXMargin,
      plotX = i MOD nCandsPerRow,
      plotY = nCandsPerCol - 1 - (i DIV nCandsPerRow),
      xCenter = plotWidth  * (FLOAT(plotX, LONG) + 0.5d0) + plotXMargin,
      yCenter = plotHeight * (FLOAT(plotY, LONG) + 0.5d0) + plotYMargin
    DO
      ps.beginDrawing(plotWidth, plotHeight, xCenter, yCenter)
    END;
  END BeginDrawing;
 
   
PROCEDURE EndDrawing(ps: PSPlot.PSFile) =
  BEGIN
    ps.endDrawing()
  END EndDrawing;
 
   
PROCEDURE DrawCandidate(
    ps: PSPlot.File;
    READONLY cand: PZCandidate.T;
    READONLY flc: ARRAY OF PZLR3Chain.ReadData;
    READONLY o: Options;
  ) =
  VAR 
    w, d: LR3.T := LR3.T{0.0d0, 0.0d0, 0.0d0};
    ini, fin: ARRAY [0..1] OF INT;
    mappedChain: ARRAY [0..1] OF REF PZLR3Chain.T;
  BEGIN
    FOR j := 0 TO 1 DO 
      IF j = 0 THEN 
        w[0]:= -1.0d0; d[1] := -o.displace/2.0d0
      ELSE 
        w[0]:= +1.0d0; d[1] := +o.displace/2.0d0
      END;
      WITH 
        s = cand.seg[j],
        chainNum = s.cvx,
        chain = flc[chainNum].c^
      DO 
        <* ASSERT NUMBER(chain) = s.tot *>
        ini[j] := s.ini MOD s.tot;
        fin[j] := (ini[j] + s.ns - 1) MOD s.tot;
        IF s.ns = 1 THEN 
          <* ASSERT s.tot >= 4 *>
          ini[j] := (ini[j]-1) MOD s.tot;
          fin[j] := (fin[j]+1) MOD s.tot
        END; 
        WITH 
          p = chain[ini[j]],
          q = chain[fin[j]],
          ctr = PZGeo.SegMid(p, q), 
          u = PZGeo.SegDir(p, ctr),
          disp = (LR3.T){-ctr[0],-ctr[1],-ctr[2]},
          t = PZGeo.Translation(disp), (* -ctr *)
          r = PZGeo.Rotation(u, w),
          s = PZGeo.Translation(d),
          m = LR4x4.Mul(LR4x4.Mul(t,r),s)
        DO
          mappedChain[j] := PZLR3Chain.Map(chain, m);
        END;     
      END;
    END;
    PZPlot.Candidate(
      ps, cand, mappedChain[0]^, mappedChain[1]^,
      whole := TRUE, closed := TRUE, 
      dots := FALSE, pairing := o.drawMatch,
      pointers := NOT o.noPointers, labels := TRUE,
      step := o.drawEvery
    )
  END DrawCandidate;

 
BEGIN
  Main()
END PZDrawCands.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
