MODULE PZGetInitialCands EXPORTS Main;

(* Generate initial candidates for multiscale matching *)
(* Last edited on 1999-11-21 02:16:15 by hcgl *)

IMPORT ParseParams, Process, FileRd, FileWr, Wr;
IMPORT OSError, Thread, Text, Stdio, Fmt;
IMPORT PZSymbolChain, PZSymbol, PZProc, PZCandidate, PZMismatch, PZSegment;

FROM Math IMPORT sqrt;
FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, INT, NAT, BOOL, BOOLS;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

CONST
  ExpectedMatchesPerCurve = 6;

TYPE
  MismatchOptions = PZMismatch.Options;

  Options = RECORD
      chainDir: TEXT;        (* Directory where to find chain data. *)
      chainPrefix: TEXT;     (* Invariant chain file name prefix. *)
      nCurves: NAT;          (* Number of curves. *)
      output: TEXT;          (* Output candidate file name. *)
      band: NAT;             (* Filtering "band" (wavelength parameter). *)
      step: LONG;            (* Sampling step. *)
      excludeSegs: TEXT;      (* Filename of segments to exclude. *)
      mp: MismatchOptions;   (* Mismatch function parameters. *)
      minLength: LONG;       (* Min length of matched segments. *)
      blurFactor: LONG;      (* Corner broadening factor. *)
      maxCurveCands: NAT;    (* Max candidates per curve. *)
      maxPairCands: NAT;     (* Max candidates per curve pair. *)
      printCands: NAT;       (* Max candidates to print. *)
      verbose: BOOL;         (* TRUE to report candidates as they are found *)
    END;

TYPE
  Segments = PZSegment.List;

  Chains = ARRAY OF PZSymbolChain.ReadData;

  VTable = PZSymbol.Table;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions(),
      band = o.band,
      lambda = FLOAT(band, LONG),
      minCandLength = EnsurePositive(o.minLength - 2.0d0 * o.blurFactor * lambda),
      sel = PZProc.SelectAll(o.nCurves)^,
      chData = PZSymbolChain.ReadAll(
        o.chainPrefix, band, sel := sel, extension := ".cvc", 
        dir := o.chainDir, headerOnly := FALSE
      ),
      excludeSegs = ReadExcludedSegsFile(o.excludeSegs)^,
      dch = PZSymbol.MakeDecodeTable(chData.sigmaMin)^,
      ech = PZSymbol.MakeErrorVarTable(chData.sigmaMin)^,
      cand = ComputeInitialCandidates(
        chData.chain^,
        dch := dch,
        ech := ech,
        mp := o.mp,
        excludeSegs := excludeSegs,
        step := o.step,
        minCandLength := minCandLength,
        maxCurveCands := o.maxCurveCands,
        maxPairCands := o.maxPairCands,
        verbose := o.verbose
      )^
    DO
      <* ASSERT chData.sigmaMin = chData.sigmaMax *>
      WITH wr = FileWr.Open(o.output & ".can") DO
        PZCandidate.Write(wr, CandComment(o), cand, lambda)
      END;
      PrintCandidates(cand, band, o.printCands);
    END;
  END Main;

PROCEDURE EnsurePositive(minCandLength: LONG): LONG =
  BEGIN
    IF minCandLength < 0.0d0 THEN
      Wr.PutText(stderr, "warning: \"-minLength\" too small for this scale\n");
      Wr.Flush(stderr);
    END;
    RETURN MAX(0.0d0, minCandLength)
  END EnsurePositive;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;

        pp.getKeyword("-nCurves");
        o.nCurves := pp.getNextInt();

        pp.getKeyword("-band");
        o.band := pp.getNextInt();

        pp.getKeyword("-step");
        o.step := pp.getNextLongReal(0.0d0);

        pp.getKeyword("-output");
        o.output := pp.getNext();

        o.mp := PZMismatch.ParseOptions(pp);

        pp.getKeyword("-minLength");
        o.minLength := pp.getNextLongReal(0.0d0);

        pp.getKeyword("-blurFactor");
        o.blurFactor := pp.getNextLongReal(0.0d0);

        IF pp.keywordPresent("-excludeSegs") THEN
          o.excludeSegs := pp.getNext()
        ELSE
          o.excludeSegs := ""
        END;

        IF pp.keywordPresent("-maxCurveCands") THEN
          o.maxCurveCands := pp.getNextInt(1)
        ELSE
          o.maxCurveCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-maxPairCands") THEN
          o.maxPairCands := pp.getNextInt(1, o.maxCurveCands)
        ELSE
          o.maxPairCands := LAST(NAT)
        END;

        IF pp.keywordPresent("-printCands") THEN
          o.printCands := pp.getNextInt(1)
        ELSE
          o.printCands := 1000
        END;

        o.verbose := pp.keywordPresent("-verbose");

        pp.finish();
      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZGetInitialCands \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix NAME \\\n");
          Wr.PutText(stderr, "  -nCurves NUMBER \\\n");
          Wr.PutText(stderr, "  -band NUMBER -step NUMBER \\\n");
          Wr.PutText(stderr, "  -output NAME \\\n");
          Wr.PutText(stderr, "  [ -excludeSegs FILENAME ] \\\n");
          Wr.PutText(stderr, "  -minLength NUMBER  -blurFactor NUMBER \\\n");
          Wr.PutText(stderr, "  " & PZMismatch.OptionsHelp & " \\\n");
          Wr.PutText(stderr, "  [ -maxCurveCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -maxPairCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -printCands NUMBER ] \\\n");
          Wr.PutText(stderr, "  [ -verbose ]\n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE ComputeInitialCandidates(
    READONLY chain: Chains;
    READONLY dch: VTable;              (* Curvature decoding table. *)
    READONLY ech: VTable;              (* Quantization noise variances. *)
    mp: MismatchOptions;               (* Mismatch formula parameters. *)
    READONLY excludeSegs: Segments;    (* List of segments to exclude *)
    step: LONG;                        (* Sampling step. *)
    minCandLength: LONG;               (* Min length of candidate *)
    maxCurveCands: NAT;                (* Max candidates per curve *)
    maxPairCands: NAT;                 (* Max candidates per curve pair *)
    verbose: BOOL;                     (* Report candidates found *)
  ): REF PZCandidate.List =
  VAR
    c : REF PZCandidate.List;
    nCands : NAT := 0;
  BEGIN
    WITH
      nChains = NUMBER(chain)
    DO
      c := NEW(REF PZCandidate.List, nChains * ExpectedMatchesPerCurve);
      FOR ka := 0 TO nChains-1 DO
        IF chain[ka].c # NIL THEN
          WITH
            a = chain[ka].c^,
            aMinSteps = FLOOR(minCandLength / step)
          DO
            FOR kb := ka+1 TO nChains-1 DO
              IF  chain[kb].c # NIL THEN
                WITH
                  b = chain[kb].c^,
                  bMinSteps = FLOOR(minCandLength / step),
                  minSteps = MIN(aMinSteps, bMinSteps)
                DO
                  Wr.PutText(stderr, ".");
                  ProcessChainPair(
                    a, ka, b, kb,
                    dch := dch,
                    ech := ech,
                    step := step,
                    minSteps := minSteps,
                    mp := mp,
                    excludeSegs := excludeSegs,
                    verbose := verbose,
                    maxPairCands := maxPairCands,
                    c := c, nCands := nCands
                  )
                END;
              END;
            END;
          END;
        END;
        Wr.PutText(stderr, " " & Fmt.Int(nCands) & "\n");
      END;
      Wr.PutText(stderr, Fmt.Int(nCands) & " candidates before curve pruning\n");
      PZCandidate.Sort(SUBARRAY(c^, 0, nCands), PZCandidate.MismatchBetter);
      PZCandidate.PruneCandsPerCurve(c^, nCands, maxCurveCands);
      Wr.PutText(stderr, Fmt.Int(nCands) & " candidates after curve pruning\n");
      PZCandidate.Sort(SUBARRAY(c^, 0, nCands), PZCandidate.PairMismatchBetter);
      WITH ct = NEW(REF PZCandidate.List, nCands) DO
        ct^ :=  SUBARRAY(c^, 0, nCands);
        RETURN ct;
      END;
    END;
  END ComputeInitialCandidates;

PROCEDURE ProcessChainPair(
    READONLY a: PZSymbolChain.T;
    ka: NAT;
    READONLY b: PZSymbolChain.T;
    kb: NAT;
    READONLY dch: VTable;              (* Curvature decoding table *)
    READONLY ech: VTable;              (* Quantization noise variances. *)
    step: LONG;                        (* Nominal step length *)
    minSteps: NAT;                     (* Minimum number of steps in candidate. *)
    mp: MismatchOptions;               (* Mismatch formula parameters. *)
    READONLY excludeSegs: Segments;    (* Segments to exclude. *)
    maxPairCands: NAT;                 (* Maximum candidates to create for this pair. *)
    verbose: BOOL;                     (* Report candidates found *)
    (*IO*)
    VAR c: REF PZCandidate.List;
    VAR nCands: NAT;
  ) =
  (*
    Consider the set of pairs "(ia,ib)" where "ia" is an index in "a"
    and "ib" is an index in "b".  Since the chains are circular, this
    set is a toroidal integer grid with periods "ma=NUMBER(a)"
    and "mb = NUMBER(b)", respectively.

    In this program, we only look for candidates that are defined
    by `perfect' matchings, i.e. matchings without half-steps.
    Such a matching is a segment of some diagonal line of the
    "(ia,ib)" grid.

    Therefore, we enumerate the diagonals of the toroidal grid, and
    search for candidates within each diagonal in turn. The number of
    diagonals is "numDiagonal = GCD(ma,mb)", and each diagonal is a
    circular sequence of "diagSize = (ma*mb)/nDiags".

    We number the diagonals from "[0..nDiags-1]"
    and their elements from [0..diagSize-1]", as follows:
    element index "r" of diagonal number "t" is
    the pair "(ia,ib)" where "ia = r MOD ma" and "ib = (t-r) MOD mb".

    Within each diagonal "t", we compute a vector "d2" such that "d2[r]"
    is the distance squared between "a[ia]" and "b[ib]", where "ia,ib"
    are the corresponding chain indices.  These distances are clipped
    TO "mp.maxDistSqr"

    We also compute a boolean vector "ok", such that "ok[r]" tells
    whether there is a sequence of "minSteps" consecutive steps
    beginning with element "r" that could be part of an acceptable
    candidate.  Specifically, "ok[r] = TRUE" iff the average squared
    differences "d2[]" over the next "minSteps" steps starting with
    pair "r" is less than "mp.critDistSqr".  Call these sequences
    `minimal candidates.'

    At the same time, we exclude from the "ok" vector any entries "ok[r]"
    such that the corresponding pair of samples, and the next "minSteps"
    samples, overlap a segment listed in "excludeSegs".

    The candidates are then found by identifying the maximal
    substrings of consecutive "TRUE"s in the "ok" vector, and checking
    whether their mismatch, plus "mp.bias", is negative.  Note that
    the average of "d2" is negative within each minimal candidate,
    hence by merging overlapping minimal candidates we only make the
    mismatch smaller.
  *)
  VAR d2Sum: LONG;
      nCandsIni: NAT := nCands;
      nCandsPair: NAT;
  BEGIN
    WITH
      ma = NUMBER(a),
      mb = NUMBER(b),
      nDiags = GCD(ma, mb),
      diagSize = (ma * mb) DIV nDiags,
      nOld = nCands + 0,
      (* Work vectors: *)
      d2 = NEW(REF LONGS, diagSize)^,      (* Squared pair distances. *)
      d2s = NEW(REF LONGS, diagSize)^,     (* Accumulated "d2[r]". *)
      ok = NEW(REF BOOLS, diagSize)^,  
      exa = NEW(REF BOOLS, ma)^,           (* Excluded parts of "a". *)
      exb = NEW(REF BOOLS, mb)^            (* Excluded parts of "b". *)
    DO
      IF verbose THEN
        Wr.PutText(stderr, "chains: " & Fmt.Int(ka) & "[" & Fmt.Int(ma) & "]");
        Wr.PutText(stderr, "  ");
        Wr.PutText(stderr, Fmt.Int(kb) & "[" & Fmt.Int(mb) & "]");
        Wr.PutText(stderr, " num diags: " & Fmt.Int(nDiags));
        Wr.PutText(stderr, " ns diag: " & Fmt.Int(diagSize));
        Wr.PutText(stderr, " minSteps: " & Fmt.Int(minSteps));
      END;

      (* Mark excluded parts of the two chains: *)
      FOR i := 0 TO LAST(exa) DO exa[i] := FALSE END;
      ExcludeSamples(ka, excludeSegs, exa, extraL := minSteps, extraR := 0);
      FOR i := 0 TO LAST(exb) DO exb[i] := FALSE END;
      ExcludeSamples(kb, excludeSegs, exb, extraL := 0, extraR := minSteps);

      (* Scan diagonals of the grid: *)
      FOR t := 0 TO nDiags-1 DO
        (*
          Compute pairwise squared distances "d2[r]".
          Also accumulates squared distances,
          so that "d2s[r] = SUM { d2[k] : k IN [0..r] }".
        *)
        d2Sum := 0.0d0;
        FOR r := 0 TO diagSize-1 DO
          WITH
            ia = r MOD ma,
            ib = (t-r) MOD mb,
            dab2 = PZSymbol.DistSqr(a[ia], b[ib], 
              complement := TRUE, decode := dch, errorVar := ech
            )
          DO
            d2[r] := MIN(mp.maxDistSqr, dab2);
            d2Sum := d2Sum + d2[r]; d2s[r] := d2Sum;
          END;
        END;
        (*
          Set "ok[r]" true if the corresponding pair "(ia,ib)"
          is the beginning of a run of "minSteps" steps whose
          average squared distance is less than "mp.critDistSqr".
        *)
        FOR r := 0 TO diagSize-1 DO
          WITH
            ia = r MOD ma,
            ib = (t - r) MOD mb,

            rLo = r,
            rmLo = rLo MOD diagSize,
            qLo = FLOAT(rLo DIV diagSize, LONG),
            s2Lo = d2s[rmLo] + qLo*d2Sum,

            rHi = r + minSteps,
            rmHi = rHi MOD diagSize,
            qHi = FLOAT(rHi DIV diagSize, LONG),
            s2Hi = d2s[rmHi] + qHi*d2Sum,

            (* Integral of "d2" by trapezoidal rule: *)
            d2Intg = (s2Hi - 0.5d0*d2[rmHi]) - (s2Lo - 0.5d0 * d2[rmLo]),
            nSteps =  FLOAT(rHi - rLo - 1, LONG)
          DO
            <* ASSERT rHi > rLo *>
            ok[r] :=
              (d2Intg < nSteps * mp.critDistSqr)
              AND (NOT exa[ia])
              AND (NOT exb[ib])
          END;
        END;

        (* Now collect runs of TRUEs in "ok": *)
        BreakDiagonal(
          a, ka,
          b, kb,
          d2, ok,
          c, nCands,
          tDiag := t,
          step := step,
          minSteps := minSteps,
          mp := mp
        );
      END;
      nCandsPair := nCands - nCandsIni;
      WITH
        cx = SUBARRAY(c^, nCandsIni, nCandsPair)
      DO
        PZCandidate.Sort(cx, PZCandidate.PairMismatchBetter);
        PZCandidate.PruneCandsPerPair(cx, nCandsPair, maxPairCands);
        nCands := nCandsIni + nCandsPair
      END;
      IF verbose THEN
        Wr.PutText(stderr, " nCands: " & Fmt.Int(nCands));
        Wr.PutText(stderr, " (+" & Fmt.Int(nCands - nOld) & ")\n");
      END;
    END;
  END ProcessChainPair;

PROCEDURE BreakDiagonal(
    READONLY a: PZSymbolChain.T;
    ka: NAT;
    READONLY b: PZSymbolChain.T;
    kb: NAT;
    READONLY d2: LONGS;
    READONLY ok: BOOLS;
    VAR c: REF  PZCandidate.List;
    VAR  nCands:  NAT;
    tDiag: NAT;              (* Index of diagonal *)
    step: LONG;
    minSteps: NAT;           (* Minimum number of steps in candidate. *)
    mp: MismatchOptions;     (* Mismatch formula parameters. *)
  ) =
  VAR
    k: NAT;
    count, kStart, kIni: NAT;
    intgDistSqr: LONG;
    nMatched: LONG;
    h: LONG;
    inx: NAT;
  BEGIN
    WITH
      ma = NUMBER(a),
      mb = NUMBER(b),
      diagSize = NUMBER(d2),
      maxSteps = MIN(ma,mb) - 1
    DO
      <* ASSERT diagSize > 1 *>
      <* ASSERT diagSize MOD ma = 0 *>
      <* ASSERT diagSize MOD mb = 0 *>
      <* ASSERT maxSteps < diagSize *>
      (* Look for a TRUE element "ok[kStart]" right after a FALSE element: *)
      k := 0;
      WHILE ok[k MOD diagSize] = TRUE AND k < diagSize DO INC(k) END;
      IF k = diagSize THEN
        (* All TRUE *)
        kStart := 0
      ELSE
        inx := 0;
        WHILE ok[k MOD diagSize] = FALSE AND inx < diagSize DO INC(k); INC(inx) END;
        IF inx = diagSize THEN (* All FALSE *) RETURN END;
        kStart := k MOD diagSize;
      END;
      (* Collect runs of TRUEs": *)
      k := kStart;
      REPEAT
        (* Here begins a new run *)
        kIni := k;
        count := 0;
        REPEAT
          INC(k);
          INC(count);
        UNTIL ok[k MOD diagSize] = FALSE OR count = diagSize;

        (* Compute end of candidate: *)
        WITH
          nStepsRaw = count - 1 + minSteps,
          nSteps = MIN(nStepsRaw, maxSteps),
          excess = nStepsRaw - nSteps
        DO
          kIni := kIni + excess DIV 2;

          IF nSteps >= minSteps THEN

            (* Make room for a new candidate: *)
            IF nCands >= NUMBER(c^) THEN
              WITH ct  = NEW(REF PZCandidate.List, nCands * 2) DO
                SUBARRAY(ct^, 0, NUMBER(c^)) := c^;
                c := ct;
              END;
            END;

            (* Save this run as a new candidate: *)
            WITH
              sa = PZSegment.T{
                cvx := ka,
                tot := ma,
                ini := kIni MOD ma,
                ns := nSteps + 1,
                rev := FALSE
              },
              sb = PZSegment.T{
                cvx := kb,
                tot := mb,
                ini := (tDiag - kIni - nSteps) MOD mb,
                ns := nSteps + 1,
                rev := TRUE
              },
              length = step * FLOAT(nSteps, LONG)
            DO
              (* Compute matched length and root mean square distance: *)
              (* Trapezoidal integration formula: *)
              intgDistSqr := 0.0d0;
              nMatched := 0.0d0;
              FOR p := 0 TO nSteps DO
                WITH r = (kIni+p) MOD diagSize DO
                  IF p = 0 OR p = nSteps THEN h := 0.5d0 ELSE h := 1.0d0 END;
                  intgDistSqr := intgDistSqr + h * d2[r];
                  IF d2[r] < mp.maxDistSqr THEN nMatched := nMatched + h END
                END
              END;
              intgDistSqr := step * intgDistSqr;
              c[nCands] := PZCandidate.T{
                seg := PZSegment.Pair{sa, sb},
                mismatch := PZMismatch.Eval(
                  length := length,
                  intgDistSqr := intgDistSqr,
                  totAsymm := 0.0d0,
                  o := mp
                ),
                length := length,
                matchedLength := step * nMatched,
                pm := NIL
              };
              INC(nCands);
            END;
          END
        END;
        WHILE ok[k MOD diagSize] = FALSE DO INC(k) END;
      UNTIL (k MOD diagSize) = kStart
    END;
  END BreakDiagonal;

PROCEDURE ExcludeSamples(
    cvx: NAT;                       (* A chain number. *)
    READONLY excludeSegs: Segments; (* Segments to exclude. *)
    VAR ex: BOOLS;                  (* OUT: samples to exclude from curve "cvx". *)
    extraL, extraR: NAT;            (* Extra samples to exclude at each END. *)
  ) =
  (*
    Sets "ex[i]" TRUE if and only if there is some excluded segment in
    "excludeSegs" that covers sample number "j" of curve "cvx", where
    "j - extraL <= i <= j + extraR". *)
  BEGIN
    FOR is := 0 TO LAST(excludeSegs) DO
      WITH s = excludeSegs[is] DO
        IF cvx = s.cvx THEN
          <* ASSERT NUMBER(ex) = s.tot *>
          FOR r := -extraL TO s.ns + extraR - 1 DO
            WITH i = (s.ini + r) MOD s.tot DO
              ex[i] := TRUE
            END
          END
        END
      END
    END
  END ExcludeSamples;

PROCEDURE GCD(a : INT; b : INT) : INT =
  VAR m : NAT := 1;
      i : NAT := 2;
  BEGIN
    WHILE  i <= MIN(a,b) DO
      IF a MOD i = 0 AND b MOD i = 0 THEN
        a := a DIV i; b := b DIV i;
        m := m * i;
        ELSE  INC(i);
      END;
    END;
    RETURN m;
  END GCD;

PROCEDURE PrintCandidates (
    READONLY cand : PZCandidate.List;
    <*UNUSED*> band: NAT;
    maxCands: NAT;
  ) =
  BEGIN
    IF maxCands = 0 THEN RETURN END;
    FOR i := 0 TO MIN(maxCands, NUMBER(cand))-1 DO
      WITH c = cand[i] DO
        Wr.PutText(stderr, "cand[" & Fmt.Int(i) & "]\n");
        PZCandidate.Print(stderr, c, pairing := FALSE)
      END
    END;
    Wr.Flush(stderr);
  END PrintCandidates;

PROCEDURE ReadExcludedSegsFile(segName: TEXT): REF Segments =
  BEGIN
    IF Text.Empty(segName) THEN
      RETURN NEW(REF Segments, 0)
    ELSE
      WITH fileName = segName & ".seg" DO
        Wr.PutText(stderr, fileName & "\n");
        WITH
          rd = FileRd.Open(fileName),
          sData = PZSegment.Read(rd)
        DO
          RETURN sData.s
        END
      END
    END
  END ReadExcludedSegsFile;

PROCEDURE CandComment(READONLY o: Options): TEXT =
  BEGIN
    RETURN
      "PZGetInitialCandidates\n" &
      "    chainDir " & o.chainDir & "\n" &
      "    chainPrefix " & o.chainPrefix & "\n" &
      "    nCurves = " & Fmt.Int(o.nCurves) & "\n" &
      "    band = " & Fmt.Int(o.band) & "\n" &
      "    maxDist = " & Fmt.LongReal(sqrt(o.mp.maxDistSqr)) & "\n" &
      "    critDist = " & Fmt.LongReal(sqrt(o.mp.critDistSqr)) & "\n" &
      "    excludeSegs = \"" & o.excludeSegs & "\"\n" &
      "    minLength = " & Fmt.LongReal(o.minLength) & "\n" &
      "    blurFactor = " & Fmt.LongReal(o.blurFactor) & "\n" &
      "    maxCurveCands = " & Fmt.Int(o.maxCurveCands) & "\n" &
      "    maxPairCands = " & Fmt.Int(o.maxPairCands) & "\n"
  END CandComment;

BEGIN
  Main()
END PZGetInitialCands.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
