MODULE PZMapCands EXPORTS Main;
(* Last edited on 1999-11-16 22:35:25 by hcgl *)

(* Maps a list of tentative matchings to another scale *)

IMPORT ParseParams, Process, FileRd, FileWr;
IMPORT Wr, OSError, Thread, Stdio, Fmt, TextWr;
IMPORT PZLRChain, PZSegment, PZCandidate;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL;

<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;        (* Input candidate file name. *)
      inBand: NAT;        (* Input nominal filtering bandwidth. *)
      chainDir: TEXT;     (* Directory where to find chain files. *)
      chainPrefix: TEXT;  (* Curve file name prefix. *)
      output: TEXT;       (* Output candidate file name. *)
      outBand: NAT;       (* Output nominal filtering bandwidth. *)
      outStep: LONG;      (* Sampling step of output curves. *)
      blurFactor: LONG;   (* Corner broadening factor. *)
      extraLength: LONG;  (* Extra length to add at each candidate end. *)

      (* Parameters that control the merging of overlapping cands: *)
      maxMergeShift: LONG;   (* Max shift btw matchings (pixels). *)
      minMergeOverlap: LONG; (* Minimum overlap btw matchings (0 = don't merge). *)
      
      verbose: BOOL;      (* TRUE to mumble while working *)
    END;

PROCEDURE Main() =
  BEGIN
    WITH
      o = GetOptions(),
      cData = PZCandidate.Read(FileRd.Open(o.input & ".can")),
      oldCand = cData.c^,
      curves = PZCandidate.GetCurves(oldCand)^,
      newLabel = PZLRChain.ReadAll(
        o.chainPrefix, o.outBand, ".lbl", sel := curves,
        dir := o.chainDir, headerOnly := FALSE
      ),
      oldLabel = PZLRChain.ReadAll(
        o.chainPrefix, o.inBand, ".lbl", sel := curves,
        dir := o.chainDir, headerOnly := FALSE
      ),
      newCand = MapAllCandidates(
        oldCand,
        oldLabel.chain^, o.inBand,
        newLabel.chain^, o.outBand,
        newStep := o.outStep, 
        blurFactor := o.blurFactor, 
        extraLength := o.extraLength, 
        tryToMerge := (o.minMergeOverlap > 0.0d0),
        maxShiftSteps := CEILING(o.maxMergeShift/o.outStep - 0.0001d0),
        minOverlapSteps := CEILING(o.minMergeOverlap/o.outStep - 0.0001d0),
        verbose := o.verbose
      )^,
      wr = FileWr.Open(o.output & ".can"),
      band = FLOAT(o.outBand,LONG)
    DO
      PZCandidate.Write(wr, cData.cmt & MapComments(o), newCand, band);
    END;
  END Main;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();

        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;

        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();

        pp.getKeyword("-inBand");
        o.inBand := pp.getNextInt();

        pp.getKeyword("-output");
        o.output := pp.getNext();

        pp.getKeyword("-outBand");
        o.outBand := pp.getNextInt();

        pp.getKeyword("-outStep");
        o.outStep := pp.getNextLongReal(0.0d0, 1024.0d0);

        pp.getKeyword("-blurFactor");
        o.blurFactor := pp.getNextLongReal(0.0d0, 5.0d0);

        pp.getKeyword("-extraLength");
        o.extraLength := pp.getNextLongReal(0.0d0);

        IF pp.keywordPresent("-mergeCands") THEN
          o.maxMergeShift := pp.getNextLongReal(0.0d0, 10000.0d0);
          o.minMergeOverlap := pp.getNextLongReal(0.0d0, 10000.0d0);
        ELSE
          o.maxMergeShift := 0.0d0;
          o.minMergeOverlap := 0.0d0;
        END;
        
        o.verbose := pp.keywordPresent("-verbose");

        pp.finish();

      EXCEPT
      | ParseParams.Error =>
          Wr.PutText(stderr, "Usage: PZMapCands \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  -inBand NUMBER \\\n");
          Wr.PutText(stderr, "  -chainPrefix NAME [ -chainDir DIR ]  \\\n");
          Wr.PutText(stderr, "  -output NAME \\\n");
          Wr.PutText(stderr, "  -outBand NUMBER \\\n");
          Wr.PutText(stderr, "  -outStep PIXELS \\\n");
          Wr.PutText(stderr, "  -blurFactor NUMBER -extraLength PIXELS \\\n");
          Wr.PutText(stderr, "  [ -mergeCands MAXSHIFT MINOVERLAP ] \\\n");
          Wr.PutText(stderr, "  [ -verbose ] \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;

PROCEDURE MapAllCandidates(
    READONLY cand: PZCandidate.List;
    READONLY oldLabel: ARRAY OF PZLRChain.ReadData;
    oldBand: NAT;
    READONLY newLabel: ARRAY OF PZLRChain.ReadData;
    newBand: NAT;
    newStep: LONG;
    blurFactor: LONG;
    extraLength: LONG;
    tryToMerge: BOOL;
    maxShiftSteps: NAT;
    minOverlapSteps: NAT;
    verbose: BOOL;
  ): REF PZCandidate.List =
  VAR
    nCandNew: NAT := 0;
  BEGIN
    WITH
      nCandOld = NUMBER(cand),
      candNew = NEW(REF PZCandidate.List, nCandOld)^
    DO
      Wr.PutText(stderr, "Candidates before mapping = " & Fmt.Int(nCandOld) & "\n");
      FOR i := 0 TO nCandOld-1  DO
        WITH
          cOld = cand[i],
          cvx0 = cOld.seg[0].cvx,
          cvx1 = cOld.seg[1].cvx,
          mappable =
            newLabel[cvx0].c # NIL AND
            newLabel[cvx1].c # NIL
        DO
          <* ASSERT oldLabel[cvx0].c # NIL AND oldLabel[cvx1].c # NIL *>
          IF verbose THEN 
            Wr.PutText(stderr, "candidate " & Fmt.Pad(Fmt.Int(i), 6, '0') & ": ");
          END;
          IF mappable THEN
            WITH
              cNew = MapCandidate(cOld,
                oldLabel, oldBand,
                newLabel, newBand,
                newStep := newStep, 
                blurFactor := blurFactor, 
                extraLength := extraLength
              )
            DO
              IF cNew.seg[0].ns > 1 AND cNew.seg[1].ns > 1 THEN 
                IF verbose THEN Wr.PutText(stderr, "  mapped! \n") END;
                candNew[nCandNew] := cNew; INC(nCandNew)
              ELSE
                IF verbose THEN Wr.PutText(stderr, "segment(s) vanished!\n") END;
              END
            END
          ELSE
            IF verbose THEN Wr.PutText(stderr, "  chain(s) missing!\n") END;
          END
        END;
      END;
      Wr.PutText(stderr, "Candidates before Sort and Merge = ");
      Wr.PutText(stderr, Fmt.Int(nCandNew) & "\n");
      PZCandidate.Sort(SUBARRAY(candNew, 0, nCandNew), PZCandidate.PairMismatchBetter);
      IF tryToMerge THEN 
        Wr.PutText(stderr, "Merging cands with at least " & Fmt.Int(minOverlapSteps) & " steps within " & Fmt.Int(maxShiftSteps) & " steps of each other\n");
        PZCandidate.MergeOverlaps(
          candNew, nCandNew, 
          minSteps := minOverlapSteps, 
          maxAdjust := maxShiftSteps
        );
        Wr.PutText(stderr, "Candidates after Sort and Merge =  ");
        Wr.PutText(stderr, Fmt.Int(nCandNew) & "\n");
      ELSE
        Wr.PutText(stderr, "Merging phase skipped\n");
      END;
      WITH r = NEW(REF PZCandidate.List, nCandNew) DO
        IF nCandNew > 0 THEN r^ := SUBARRAY(candNew, 0, nCandNew) END;
        RETURN r;
      END;
    END;
  END MapAllCandidates;

PROCEDURE MapCandidate(
    READONLY cOld: PZCandidate.T;
    READONLY oldLabel: ARRAY OF PZLRChain.ReadData;
    oldBand: NAT;
    READONLY newLabel: ARRAY OF PZLRChain.ReadData;
    newBand: NAT;
    newStep: LONG;
    blurFactor: LONG;
    extraLength: LONG;
  ): PZCandidate.T =
  VAR
    seg: PZSegment.Pair;
    tOld: ARRAY [0..1] OF REF PZLRChain.T;
    tNew: ARRAY [0..1] OF REF PZLRChain.T;
    stride: ARRAY [0..1] OF LONG;
    stepNew: ARRAY [0..1] OF LONG;
    extra: ARRAY [0..1] OF INT;
  BEGIN
    FOR j := 0 TO 1 DO
      WITH
        s = cOld.seg[j],
        halfShrink = blurFactor * FLOAT(newBand - oldBand, LONG) / newStep,
        halfStretch = extraLength / newStep
      DO
        seg[j] := s;
        tOld[j] := oldLabel[s.cvx].c;
        tNew[j] := newLabel[s.cvx].c;
        stride[j] := oldLabel[s.cvx].stride;
        stepNew[j] := newStep;
        extra[j] := ROUND(halfStretch - halfShrink);
        (* Wr.PutText(stderr, "  label stride = " & Fmt.LongReal(stride[j]) & "\n"); *)
        <* ASSERT oldLabel[s.cvx].stride = newLabel[s.cvx].stride *>
      END;
    END;
    WITH
      mapCand =  PZLRChain.MapCandidate(
        cOld, tOld[0]^, tOld[1]^, tNew[0]^, tNew[1]^, stride, stepNew, extra
      )
    DO
      RETURN  mapCand
    END
  END MapCandidate;

PROCEDURE MapComments(READONLY o: Options):TEXT=
  BEGIN
    WITH
      wr = NEW(TextWr.T).init()
    DO
      Wr.PutText( wr, "PZMapCands \n");
      Wr.PutText( wr, "  input =           \"" & o.input & "\"\n");
      Wr.PutText( wr, "  inBand =          " & Fmt.Int(o.inBand) & "\n");
      Wr.PutText( wr, "  chainDir =        \"" & o.chainDir     & "\"\n");
      Wr.PutText( wr, "  chainPrefix =     \"" & o.chainPrefix     & "\"\n");
      Wr.PutText( wr, "  output =          \"" & o.output & "\"\n");
      Wr.PutText( wr, "  outBand =         " & Fmt.Int(o.outBand) & "\n");
      Wr.PutText( wr, "  blurFactor =      " & FLR(o.blurFactor,8,6) & "\n");
      Wr.PutText( wr, "  extraLength =     " & FLR(o.extraLength,8,4) & "\n");
      Wr.PutText( wr, "  outStep =         " & FLR(o.outStep,8,4) & "\n");
      Wr.PutText( wr, "  maxMergeShift =   " & FLR(o.maxMergeShift,8,4) & "\n");
      Wr.PutText( wr, "  minMergeOverlap = " & FLR(o.minMergeOverlap,8,4) & "\n");
      RETURN(TextWr.ToText(wr))
    END (* DO *);
  END MapComments;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

BEGIN
  Main()
END PZMapCands.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
