MODULE PZOptMatrix;

(* Last edited on 1999-09-04 01:12:42 by hcgl *)

IMPORT FileWr, Text, Wr, Fmt, OSError, Thread, Math;
IMPORT LR4x4, LR3, Minimizer, CoordMinimizer, BrentUniMin;
IMPORT PZGeo;

FROM Minimizer IMPORT Point, Gradient, Value;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL;
FROM Stdio IMPORT stderr;


PROCEDURE Optimize(
    func: MatrixFunction;  (* Function to optimize. *)
    scaleX, scaleY: LONG;  (* Scale lengths for translation. *)
    maxEvals: NAT;         (* Maximum number of "func" calls. *) 
    VAR nEvals: NAT;       (* Total number of "func" calls performed. *)
    VAR m: LR4x4.T;        (* IN: initial guess, OUT: optimized matrix. *)
    VAR f: LONG;           (* OUT: "func(m)". *)
    plotName: TEXT;        (* Plot file for "func" (minus extension), or empty. *)
  ) = 
  BEGIN
    WITH 
      uMin = NEW(BrentUniMin.T),
      cMin = NEW(CoordMinimizer.T).setUniMin(uMin).setBudget(5).setSize(3), 
      g    = NEW(REF Gradient, 3)^,
      x    = NEW(REF Point, 3)^,
      xIni = NEW(REF Point, 3)^
    DO

      PROCEDURE ConvertXtoM(READONLY x: Point; VAR m : LR4x4.T)=
        BEGIN 
          m := PZGeo.Translation(LR3.T{scaleX*x[0], scaleY*x[1], 0.0d0});
          m[1, 1] := Math.cos(x[2]);
          m[1, 2] := Math.sin(x[2]);
          m[2, 1] := -m[1, 2];
          m[2, 2] := m[1, 1];
        END ConvertXtoM;
   
      PROCEDURE ConvertMtoX(READONLY  m : LR4x4.T; VAR  x: Point)=
        BEGIN 
          <* ASSERT scaleX # 0.0d0  *>
          <* ASSERT scaleY # 0.0d0  *>
          x[0] := m[0, 1]/scaleX;
          x[1] := m[0, 2]/scaleY;
          x[2] := Math.atan2(m[1, 2], m[1, 1]);
        END  ConvertMtoX;
        
      PROCEDURE Evaluate(
         READONLY x: Point; 
         VAR f: Value; 
         <*UNUSED*> VAR g: Gradient
        ): BOOL =
        <* FATAL Wr.Failure, Thread.Alerted *>
        VAR m :  LR4x4.T;
        BEGIN
          IF nEvals >= maxEvals THEN RETURN TRUE END;
          INC(nEvals);
          Wr.PutText(stderr, "evals = " & FI(nEvals,6) & " trn = (" & FLR(x[0],8,5) & ", " & FLR(x[1],8,5) & ") rot = " & FLR(x[2],8,5));
          ConvertXtoM(x, m);
          f := func(m);
          Wr.PutText(stderr, " val = " & FLR(f,12,4) & "\n");
          RETURN FALSE;
        END Evaluate;
  
      PROCEDURE PlotEvalSection(READONLY u, v: Point)=
        VAR f : Value;
        CONST NSteps = 50;
        VAR saveMaxEvals: NAT := maxEvals;
        <* FATAL OSError.E, Wr.Failure, Thread.Alerted *>
        BEGIN
          WITH 
            wrFile = FileWr.Open(plotName & ".plt"),
            g =  NEW(REF Gradient, 3)^,
            y = NEW(REF Point, 3)^
          DO 
            maxEvals := LAST(NAT);
            FOR i := 0 TO 2*NSteps DO 
              WITH 
                s = FLOAT(i,LONG)/FLOAT(NSteps, LONG),
                t = 1.0d0 - s
              DO
                y[0] := t*u[0] + s*v[0];
                y[1] := t*u[1] + s*v[1];
                y[2] := t*u[2] + s*v[2];
                EVAL Evaluate(y, f, g);
                Wr.PutText(wrFile, Fmt.LongReal(t)); Wr.PutText(wrFile, "  ");
                Wr.PutText(wrFile, Fmt.LongReal(y[0])); Wr.PutText(wrFile, "  ");
                Wr.PutText(wrFile, Fmt.LongReal(y[1])); Wr.PutText(wrFile, "  ");
                Wr.PutText(wrFile, Fmt.LongReal(y[2])); Wr.PutText(wrFile, "  ");
                Wr.PutText(wrFile, Fmt.LongReal(f)); Wr.PutText(wrFile, "  ");
                Wr.PutText(wrFile, "\n");
              END;
            END;
            Wr.Close(wrFile);
            maxEvals := saveMaxEvals;
          END;  
       END PlotEvalSection;
     
      BEGIN 
        ConvertMtoX(m,x);
        EVAL Evaluate(x, f, g);
        xIni := x;
        
        cMin.debug := FALSE; 
        cMin.minimize(
          eval := Evaluate,      (* Evaluates the goal function. *)
          x := x,                (* In: initial guess, out: minimum point. *) 
          fx := f,               (* In/out: function value at "x" *)
          dfx := g,              (* In/out: gradient of "f" at "x" *)
          dist := 1.0d-1,        (* Guessed distance from optimum *)
          tol := 1.0d-1/scaleY,  (* Guessed radius of optimum region *)
          check := NIL           (* Acceptance test *)    
        );

        EVAL Evaluate(x, f, g);
        ConvertXtoM(x,m);
        
        IF NOT Text.Empty(plotName) THEN
          PlotEvalSection(xIni, x)
        END;
      END;
    END;
  END Optimize;


PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

PROCEDURE FI(x: INT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

BEGIN
END PZOptMatrix.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
